package backend

import (
	"context"
	"errors"
	"io/ioutil"
	"log"
	"strings"

	ldaputil "git.autistici.org/ai3/go-common/ldap"
	"gopkg.in/ldap.v2"
)

type LDAPQueryConfig struct {
	// SearchBase, SearchFilter and Scope define parameters for
	// the LDAP search. The search should return a single object.
	// SearchBase and SearchFilter can contain the string "%s",
	// which will be replaced with the username before performing
	// a query.
	SearchBase   string `yaml:"search_base"`
	SearchFilter string `yaml:"search_filter"`
	ScopeStr     string `yaml:"scope"`
	scope        int

	// Attr is the LDAP attribute holding the encrypted user key.
	PublicKeyAttr  string `yaml:"attr"`
	PrivateKeyAttr string `yaml:"attr"`
}

// Valid returns an error if the configuration is invalid.
func (c *LDAPQueryConfig) Valid() error {
	if c.SearchBase == "" {
		return errors.New("empty search_base")
	}
	if c.SearchFilter == "" {
		return errors.New("empty search_filter")
	}
	c.scope = ldap.ScopeWholeSubtree
	if c.ScopeStr != "" {
		s, err := ldaputil.ParseScope(c.ScopeStr)
		if err != nil {
			return err
		}
		c.scope = s
	}
	if c.PublicKeyAttr == "" {
		return errors.New("empty public_key_attr")
	}
	if c.PrivateKeyAttr == "" {
		return errors.New("empty public_key_attr")
	}
	return nil
}

func (c *LDAPQueryConfig) searchRequest(username string, attrs ...string) *ldap.SearchRequest {
	u := ldap.EscapeFilter(username)
	base := strings.Replace(c.SearchBase, "%s", u, -1)
	filter := strings.Replace(c.SearchFilter, "%s", u, -1)
	return ldap.NewSearchRequest(
		base,
		c.scope,
		ldap.NeverDerefAliases,
		0,
		0,
		false,
		filter,
		attrs,
		nil,
	)
}

// LDAPConfig holds the global configuration for the LDAP user backend.
type LDAPConfig struct {
	URI        string           `yaml:"uri"`
	BindDN     string           `yaml:"bind_dn"`
	BindPwFile string           `yaml:"bind_pw_file"`
	Query      *LDAPQueryConfig `yaml:"query"`
}

// Valid returns an error if the configuration is invalid.
func (c *LDAPConfig) Valid() error {
	if c.URI == "" {
		return errors.New("empty uri")
	}
	if c.BindDN == "" {
		return errors.New("empty bind_dn")
	}
	if c.BindPwFile == "" {
		return errors.New("empty bind_pw_file")
	}
	if c.Query == nil {
		return errors.New("missing query configuration")
	}
	return c.Query.Valid()
}

type ldapBackend struct {
	config *LDAPConfig
	pool   *ldaputil.ConnectionPool
}

func NewLDAPBackend(config *LDAPConfig) (*ldapBackend, error) {
	// Validate configuration.
	if err := config.Valid(); err != nil {
		return nil, err
	}

	// Read the bind password.
	bindPw, err := ioutil.ReadFile(config.BindPwFile)
	if err != nil {
		return nil, err
	}

	// Connect.
	pool, err := ldaputil.NewConnectionPool(config.URI, config.BindDN, strings.TrimSpace(string(bindPw)), 5)
	if err != nil {
		return nil, err
	}

	return &ldapBackend{
		config: config,
		pool:   pool,
	}, nil
}

func (b *ldapBackend) GetPrivateKeys(ctx context.Context, username string) [][]byte {
	result, err := b.pool.Search(ctx, b.config.Query.searchRequest(username, b.config.Query.PrivateKeyAttr))
	if err != nil {
		log.Printf("LDAP error: %v", err)
		return nil
	}

	var out [][]byte
	for _, ent := range result.Entries {
		for _, val := range ent.GetAttributeValues(b.config.Query.PrivateKeyAttr) {
			out = append(out, []byte(val))
		}
	}
	return out
}

func (b *ldapBackend) GetPublicKey(ctx context.Context, username string) []byte {
	result, err := b.pool.Search(ctx, b.config.Query.searchRequest(username, b.config.Query.PublicKeyAttr))
	if err != nil {
		log.Printf("LDAP error: %v", err)
		return nil
	}
	if len(result.Entries) == 0 {
		return nil
	}
	if len(result.Entries) > 1 {
		log.Printf("public key query for %s returned too many results (%d)", username, len(result.Entries))
		return nil
	}

	s := result.Entries[0].GetAttributeValue(b.config.Query.PublicKeyAttr)
	if s == "" {
		return nil
	}
	return []byte(s)
}
